@@ -9,7 +9,7 @@
 
 require 5.004;
 
-$DBD::ODBC::VERSION = '0.42';
+$DBD::ODBC::VERSION = '0.43';
 
 {
     package DBD::ODBC;
@@ -102,10 +102,29 @@ $DBD::ODBC::VERSION = '0.42';
 	$sth;
     }
 
+    sub column_info {
+	my ($dbh, $catalog, $schema, $table, $column) = @_;
 
+	$catalog = "" if (!$catalog);
+	$schema = "" if (!$schema);
+	$table = "" if (!$table);
+	$column = "" if (!$column);
+	# create a "blank" statement handle
+	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLColumns" });
+
+	_columns($dbh,$sth, $catalog, $schema, $table, $column)
+	    or return undef;
+
+	$sth;
+    }
+    
     sub columns {
 	my ($dbh, $catalog, $schema, $table, $column) = @_;
 
+	$catalog = "" if (!$catalog);
+	$schema = "" if (!$schema);
+	$table = "" if (!$table);
+	$column = "" if (!$column);
 	# create a "blank" statement handle
 	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLColumns" });
 
@@ -416,6 +435,23 @@ See L<DBI> for more information.
  t/09multi.t, if your driver doesn't seem to support
  returning multiple result sets.
 
+=item B<DBD::ODBC 0.43>
+
+Fix for FoxPro drivers -- HURRAY!!!!!!!!!  Bad call to SQLDescribeCol where the FoxPro driver
+seems to clear out the buffer passed and the size of the buffer indicated in the parameters
+was larger than the real buffer...need better fix long term, but fixed for now by allocating
+extra space.
+ 
+Added support for column_info (maps to function columns, previously provided)
+ 
+Fix for binding undef value generated by a dereferencing an empty hash value.
+I.e in the hash, there is no key value "1", but $hash{"1"} is passed to binding parameters.
+
+A fix to make all bound columns word aligned.  This solved a problem on solaris when binding
+timestamp values after a varchar value.  This patch forces all columns to start on word
+boundaries (sizeof int boundaries, actually).  This may have a positive speed impact for
+other systems.  Thanks to Joe Slagel for pointing this out and providing a patch for the fix!
+ 
 =item B<DBD::ODBC 0.42>
 
 A patch to the tests to get past any potential problems with
